# Copyright 2013-2022 Rumma & Ko Ltd.
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""
Defines tools for testing.

- The :func:`make_docs_suite` function to automatically run :cmd:`doctest` on each document of a doctree.
- an extended TestCase with methods to launch a subprocess.

"""

import os
from os.path import join, pathsep
import unittest
import glob
from fnmatch import fnmatch
import sys
from setuptools import find_packages
# from pathlib import Path
import subprocess

from atelier.utils import SubProcessParent

DOCTEST_CMD = ["doctest", "-o", "REPORT_ONLY_FIRST_FAILURE"]


def interpreter_args():
    # if 'coverage' in sys.modules:
    #     # raise Exception('20160119')
    #     return ['coverage', 'run']
    return [sys.executable]


class DocTestCase(unittest.FunctionTestCase, SubProcessParent):
    # internally used by make_docs_suite

    def __init__(self, filename, addenv=None):
        self.addenv = addenv

        def func():
            args = [sys.executable]
            args += ["-m"]
            args += DOCTEST_CMD
            args += [filename]
            self.run_subprocess(args)

        func.__name__ = filename
        super(DocTestCase, self).__init__(func)

    def build_environment(self):
        env = super(DocTestCase, self).build_environment()
        env.pop('PYTHONPATH', None)  # fixes #1296
        if self.addenv is not None:
            env.update(self.addenv)
        return env


def make_docs_suite(docs_root, include="*.rst", exclude=None, addenv=None):
    """
    Discover the doc files in specified directory docs_root and below
    and return a test suite that tests them all, each one in a
    separate subprocess.

    `include` is a filename pattern of the files to include. Default
    is `'*.rst'`.

    `exclude` is an optional filename pattern of the files to exclude. It can be
    a `str` with a single pattern or multiple patterns separated by `:`, or an
    iterator that yields patterns. If a filename matches any of the patterns, it
    will be skipped. Default value is None.

    `addenv` is an optional dictionary with additional environment
    variables to be set in the subprocess. Default is None.

    The tests are sorted alphabeticallly in order to avoid surprises
    when some doctest inadvertantly modifies the database.
    """
    suite = unittest.TestSuite()
    count = 0
    if exclude is None:
        exclude_patterns = []
    elif isinstance(exclude, str):
        exclude_patterns = exclude.split(pathsep)
    else:
        exclude_patterns = exclude

    def match_exclude(fn):
        for pat in exclude_patterns:
            if fnmatch(fn, pat):
                return True
        return False

    for root, dirs, files in os.walk(docs_root):
        dirs.sort()
        for file in sorted(files):
            fn = join(root, file)
            if fnmatch(fn, include):
                if match_exclude(fn):
                    print("Skip doctest for {}".format(fn))
                    continue
                suite.addTest(DocTestCase(fn, addenv))
                count += 1
    print("Loaded {} doctests from {}".format(count, docs_root))
    return suite


class TestCase(unittest.TestCase, SubProcessParent):
    "A unittest testcase with some additional utility methods."

    project_root = NotImplementedError

    # maxDiff = None

    def run_packages_test(self, declared_packages):
        """
        Checks whether the `packages` parameter to setup seems correct, i.e.
        whether all packages in the repository are being published in a source
        distribution.  This is kind of basic hygiene because you usually forget
        to update this parameter when you add or remove a package.

        """
        found_packages = find_packages()
        # if tests exists, remove it:
        if 'tests' in found_packages:
            found_packages.remove('tests')
        found_packages.sort()
        declared_packages.sort()
        self.assertEqual(found_packages, declared_packages)

    def run_simple_doctests(self, filenames, **kw):  # env.simple_doctests
        """
        run doctest of given file in a subprocess
        """
        for ln in filenames.splitlines():
            ln = ln.strip()
            if ln and not ln.startswith('#'):
                ok = False
                for fn in glob.glob(ln):
                    ok = True
                    args = interpreter_args()
                    args += ["-m"]
                    # args += ["doctest"]
                    args += DOCTEST_CMD
                    args += [fn]
                    self.run_subprocess(args, **kw)
                if not ok:
                    self.fail("no files matching %s" % ln)

    def run_unittest(self, filename, **kw):
        """
        run unittest of given file in a subprocess
        """
        args = interpreter_args()
        #~ args += ["-Wall"]
        args += ["-m"]
        args += ["unittest"]
        #~ args += ["--buffer"]
        args += [filename]
        self.run_subprocess(args, **kw)



# -*- coding: UTF-8 -*-
# Copyright 2016-2020 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""
A library of `invoke
<http://docs.pyinvoke.org/en/latest/index.html>`__ tasks.  See
:doc:`/invlib`.

.. autosummary::
   :toctree:

   tasks
   utils
"""

import os

from importlib import import_module
from invoke import Collection
from pathlib import Path

import atelier


def setup_from_tasks(globals_dict,
                     main_package=None,
                     django_settings_module=None,
                     **kwargs):
    """
    This is the function you must call from your :xfile:`tasks.py` file
    in order to activate the tasks defined by atelier.

    Arguments:

    - `globals_dict` must be the `globals()` of the calling script.

    - Optional `main_package` is the name of the main Python package provided by
      this project.

    - Optional `django_settings_module` will be stored in the
      :envvar:`DJANGO_SETTINGS_MODULE`, and certain project configuration
      options will get their default value from that module.

    - All remaining keyword arguments are project configuration parameters and
      stored to the :ref:`project configuration options <atelier.prjconf>`.

    """
    if '__file__' not in globals_dict:
        raise Exception(
            "No '__file__' in %r. "
            "First argument to setup_from_tasks() must be `globals()`." %
            globals_dict)

    tasks_file = Path(globals_dict['__file__'])
    if not tasks_file.exists():
        raise Exception("No such file: %s" % tasks_file)
    # print("20180428 setup_from_tasks() : {}".format(root_dir))

    from atelier.invlib import tasks
    from atelier.projects import get_project_from_path
    prj = get_project_from_path(tasks_file.parent)

    if atelier.current_project is None:
        atelier.current_project = prj

    if kwargs:
        for k, v in kwargs.items():
            if k not in prj.config:
                raise Exception("Invalid keyword argument '{}={}'.".format(
                    k, v))
        prj.config.update(kwargs)

    if django_settings_module is not None:
        os.environ['DJANGO_SETTINGS_MODULE'] = django_settings_module
        from django.conf import settings
        prj.config.update(
            languages=[lng.name for lng in settings.SITE.languages])

    if isinstance(main_package, str):
        main_package = import_module(main_package)
    if main_package:
        prj.set_main_package(main_package)

    self = Collection.from_module(tasks)
    prj.set_namespace(self)

    return self



# Copyright 2011-2022 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""
See :doc:`/index`.

.. autosummary::
   :toctree:

   invlib
   invlib.utils
   invlib.tasks
   jarbuilder
   projects
   test
   utils
   sheller

"""

import setuptools  # try to avoid "Distutils was imported before Setuptools"

from .setup_info import SETUP_INFO

__version__ = SETUP_INFO['version']

# intersphinx_urls = dict(docs="https://lino-framework.gitlab.io/atelier")
intersphinx_urls = dict(docs="https://atelier.lino-framework.org")
srcref_url = 'https://gitlab.com/lino-framework/atelier/blob/master/%s'
# doc_trees = ['docs']

current_project = None
"""
The currently loaded project.  An instance of
:class:`atelier.Project`.  This is set by :func:`atelier.invlib.setup_from_tasks`.
"""



# Copyright 2013-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

# This module has no docstring because it is to be execfile'd
# from `setup.py`, `atelier/__init__.py` and possibly some external
# tools, too.

SETUP_INFO = dict(name='atelier',
                  version='1.1.53',
                  description="A collection of tools for software artists",
                  license_files=['COPYING'],
                  author='Rumma & Ko Ltd',
                  author_email='info@lino-framework.org',
                  url="https://gitlab.com/lino-framework/atelier")

install_requires = [
    'invoke', 'argh', 'six', 'future', 'Babel', 'python-dateutil', 'Sphinx',
    'rstgen'
]
install_requires.append('requests')
install_requires.append('sphinx-design')
# install_requires.append('sphinx-rtd-theme')
install_requires.append('pydata-sphinx-theme')
install_requires.append('insipid-sphinx-theme')
install_requires.append('gitpython')

# trying to fix #3246
tests_require = ['pytest-cov']

# # Explicitly install `importlib` under Python 2.6. Thanks to
# # https://stackoverflow.com/questions/9418064
# try:
#     import importlib
# except ImportError:
#     install_requires.append('importlib')

SETUP_INFO.update(
    install_requires=install_requires,
    tests_require=tests_require,
    # scripts=['scripts/per_project'],
    entry_points={
        'console_scripts': ['per_project = atelier.cli.per_project:main']
    },
    test_suite='tests',
    long_description="""\

`atelier` is a collection of tools for maintaining your Python software
repositories using a library of `invoke <https://www.pyinvoke.org/>`_ commands.

- Source code: https://gitlab.com/lino-framework/atelier
- Documentation: https://lino-framework.gitlab.io/atelier

""",
    classifiers="""\
Programming Language :: Python
Programming Language :: Python :: 3.7
Framework :: Sphinx :: Extension
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
License :: OSI Approved :: GNU Affero General Public License v3
Natural Language :: English
Operating System :: OS Independent""".splitlines())

SETUP_INFO.update(packages=[
    n for n in """
atelier
atelier.invlib
atelier.cli
""".splitlines() if n
])

# SETUP_INFO.update(
#     zip_safe=False,
#     include_package_data=True)
#
# SETUP_INFO.update(package_data=dict())
#
# def add_package_data(package, *patterns):
#     l = SETUP_INFO['package_data'].setdefault(package, [])
#     l.extend(patterns)
#     return l
#
# add_package_data('rstgen.sphinxconf', '*.html')



# -*- coding: UTF-8 -*-
# Copyright 2016-2018 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
raise Exception("""
Backwards-incompatible new syntax for :xfile:`tasks.py` files.

Before::

    from atelier.invlib.ns import ns
    ns.setup_from_tasks(globals(), ...)

After::

    from atelier.invlib import setup_from_tasks
    ns = setup_from_tasks(globals(), ...)


""")
# from atelier.invlib import MyCollection
# from . import tasks
#
# ns = MyCollection.from_module(tasks)



# -*- coding: UTF-8 -*-
# Copyright 2013-2022 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""
This is the module that defines the invoke namespace.

It is imported by :func:`atelier.invlib.setup_from_tasks` which passes
it to :func:`invoke.Collection.from_module`.
"""

from os.path import commonprefix
import os
import sys
import io
import glob
import time
import datetime
import shutil
from datetime import timedelta
from importlib import import_module

from atelier.utils import i2d, i2t
from babel.dates import format_date
import rstgen
from atelier.projects import load_projects
from pathlib import Path

try:
    from invoke import ctask as task  # , tasks
    # before version 0.13 (see http://www.pyinvoke.org/changelog.html)
except ImportError:
    from invoke import task  # , tasks

from invoke.exceptions import Exit
from invoke import run

import atelier
from atelier.utils import confirm, cd

from .utils import must_confirm

import requests


def checkurl(url, name):
    r = requests.get(url, allow_redirects=False)
    if r.status_code != 200:
        print("{name} ({url}) returns {r.status_code}".format(**locals()))


def add_slash(url):
    if not url.endswith('/'):
        url += "/"
    return url


LASTREL_INFO = "Last PyPI release was %(filename)s \
(%(upload_time)s,  %(downloads)d downloads)."

RELEASE_CONFIRM = """
This is going to officially release %(name)s %(version)s to PyPI.
It will fail if version %(version)s of %(name)s has previously been released.
Your `docs/changes.rst` should have a section about this version.
Your working directory should be clean (otherwise answer 'n' and run `inv ci`).
Are you sure?"""


def local(*args, **kwargs):  # probably no longer used
    """Call :func:`invoke.run` with `pty=True
    <http://www.pyinvoke.org/faq.html#why-is-my-command-behaving-differently-under-invoke-versus-being-run-by-hand>`_.

    This is useful e.g. to get colors in a terminal.

    """
    kwargs.update(pty=True)
    # kwargs.update(encoding='utf-8')
    run(*args, **kwargs)


def get_current_date(today=None):
    if today is None:
        # return datetime.datetime.utcnow()
        return (datetime.date.today(), datetime.datetime.now())
    parts = today.split(":")
    if len(parts) == 1:
        return (i2d(today), datetime.datetime.now())
    return (i2d(parts[0]), i2t(parts[1]))


def set_times(pth):
    """Set the timestamps of the specified file to now.

    Same effect as running :cmd:`touch <pth>` in a shell.

    """
    mtime = time.time()
    try:
        os.utime(pth, (mtime, mtime))
    except FileNotFoundError:
        raise FileNotFoundError("No such file: {}".format(pth))


def rmtree_after_confirm(p, batch=False):
    if not p.exists():
        # print("20210425 does not exist:", str(p))
        return
    what = "Remove %s and everything under it" % p.absolute()
    if batch or confirm("OK to {}?".format(what)):
        if batch:
            print(what)
        shutil.rmtree(p)


def cleanup_pyc(p, batch=False):  # no longer used
    """Thanks to oddthinking on https://stackoverflow.com/questions/2528283
    """
    for root, dirs, files in os.walk(p):
        pyc_files = [
            filename for filename in files if filename.endswith(".pyc")
        ]
        py_files = set(
            [filename for filename in files if filename.endswith(".py")])
        excess_pyc_files = [
            pyc_filename for pyc_filename in pyc_files
            if pyc_filename[:-1] not in py_files
        ]
        for excess_pyc_file in excess_pyc_files:
            full_path = os.path.join(root, excess_pyc_file)
            if batch or confirm("Remove excess file %s:" % full_path):
                os.remove(full_path)


def sphinx_clean(ctx, batch=False):
    """Delete all generated Sphinx files.

    """
    for b in atelier.current_project.get_doc_trees():
        if b.src_path:
            rmtree_after_confirm(b.src_path / '.doctrees', batch)
            rmtree_after_confirm(b.out_path, batch)


def py_clean(ctx, batch=False):
    """
    Delete :xfile:`.pyc` files, :xfile:`.eggs` and :xfile:`__cache__`
    directories under the project's root direcotory.
    """
    paths = []
    for root, dirs, files in os.walk(ctx.root_dir):
        p = Path(root) / '__pycache__'
        if p.exists():
            paths.append(p)
    if len(paths):
        if batch or confirm("Remove {0} __pycache__ directories".format(
                len(paths))):
            for p in paths:
                rmtree_after_confirm(p, True)

    for root, dirs, files in os.walk(ctx.root_dir):
        for fn in files:
            if fn.endswith(".pyc"):
                full_path = os.path.join(root, fn)
                if batch or confirm("Remove file %s:" % full_path):
                    os.remove(full_path)

    # cleanup_pyc(ctx.root_dir, batch)

    # if atelier.current_project.main_package is not None:
    #     try:
    #         p = Path(atelier.current_project.main_package.__file__).parent
    #         cleanup_pyc(atelier.current_project.root_dir, batch)
    #     except AttributeError:
    #         # happened 20170310 in namespace package:
    #         # $ pywhich commondata
    #         # Traceback (most recent call last):
    #         #   File "<string>", line 1, in <module>
    #         # AttributeError: 'module' object has no attribute '__file__'
    #         pass

    p = ctx.root_dir / '.eggs'
    if p.exists():
        rmtree_after_confirm(p, batch)

    files = []
    for pat in ctx.cleanable_files:
        for p in glob.glob(os.path.join(ctx.root_dir, pat)):
            files.append(p)
    if len(files):
        if batch or confirm("Remove {0} cleanable files".format(len(files))):
            for p in files:
                os.remove(p)


class RstFile(object):

    def __init__(self, local_root, url_root, parts):
        self.path = local_root / ('/'.join(parts) + '.rst')
        self.url = url_root + "/" + "/".join(parts) + '.html'
        # if parts[0] == 'docs':
        #     self.url = url_root + "/" + "/".join(parts[1:]) + '.html'
        # else:
        #     raise Exception("20131125")
        # self.url = url_root + "/" + "/".join(parts) + '.html'


class MissingConfig(Exception):

    def __init__(self, name):
        msg = "Must set `config.{0}` in `tasks.py`!"
        msg = msg.format(name)
        Exception.__init__(self, msg)


@task(name='updatefixtures')
def update_fixtures(ctx):
    # print(ctx.languages)
    if ctx.fixtures_updater:
        ctx.fixtures_updater(ctx)
    # m = import_module(ctx.fixtures_updater)
    # m.update_fixtures(ctx)


@task(name='run')
def run_script(ctx, script, after=None, start=None):
    """Run a given python script on all the demo projects."""

    cmd = f"manage.py run {Path(script).resolve()}"
    run_in_demo_projects(ctx, cmd, after=after, start=start)


@task(name='test')
def run_tests(ctx):
    """Run the test suite of this project."""
    # assert os.environ['COVERAGE_PROCESS_START']
    cmd = ctx.test_command
    if cmd:
        print("Run test command {0} in {1}:".format(cmd, ctx.root_dir))
        with ctx.cd(ctx.root_dir):
            ctx.run(cmd, pty=True)

    # if (ctx.root_dir / 'tox.ini').exists():
    #     ctx.run("REQ_VERSION=local tox", pty=True)
    # elif (ctx.root_dir / 'setup.py').exists():
    #     ctx.run(sys.executable + ' setup.py -q test', pty=True)


@task(name='readme')
def write_readme(ctx):
    """Generate or update `README.txt` or `README.rst` file from `SETUP_INFO`. """
    if not atelier.current_project.main_package:
        return
    atelier.current_project.load_info()
    info = atelier.current_project.SETUP_INFO
    long_description = info.get('long_description')
    if not long_description:
        return
    if long_description.startswith("===="):  # e.g. in synodal
        return
    # if len(ctx.doc_trees) == 0:
    #     # when there are no docs, then the README file is manually maintained
    #     return
    if ctx.revision_control_system == 'git':
        readme = ctx.root_dir / 'README.rst'
    else:
        readme = ctx.root_dir / 'README.txt'

    # for k in ('name', 'description', 'long_description', 'url'):
    #     if k not in env.current_project.SETUP_INFO:
    #         msg = "SETUP_INFO for {0} has no key '{1}'"
    #         raise Exception(msg.format(env.current_project, k))

    title = rstgen.header(1, "The ``{}`` package".format(info['name']))

    do_not_edit = """
..
    This file has been generated by ``inv bd` from the docstring of the
    :file:`setup.py` file. Do not edit this file, but edit the :file:`setup.py` file
    and run ``doctest setup.py`` before generating it again.

"""
    do_not_edit = "\n"  # good idea, but lets do it later

    txt = """\
{title}{do_not_edit}
{long_description}
""".format(title=title, do_not_edit=do_not_edit, **info)
    if readme.exists() and readme.read_text() == txt:
        return
    must_confirm("Overwrite %s" % readme.absolute())
    readme.write_text(txt)
    docs_index = ctx.root_dir / 'docs' / 'index.rst'
    if docs_index.exists():
        mtime = time.time()
        os.utime(docs_index, (mtime, mtime))


@task(write_readme, name='bd')
def build_docs(ctx, only=None):
    """Build docs. Build all Sphinx HTML doctrees for this project. """
    # print("Build docs for {}".format(atelier.current_project))

    cmd = ctx.make_docs_command
    if cmd:
        print("Invoking make_docs_command {0}".format(cmd))
        ctx.run(cmd, pty=True)
    for tree in atelier.current_project.get_doc_trees():
        if tree.src_path:
            if only is None or tree.rel_path == only:
                tree.build_docs(ctx)


@task(name='clean')
def clean(ctx, batch=False):
    """Remove temporary and generated files."""
    # def clean(ctx, *cmdline_args):
    sphinx_clean(ctx, batch)
    py_clean(ctx, batch)
    # clean_demo_caches()


@task(name='pepper')
def pepper(ctx):
    "Run PEP formatting and checking."
    # NB The ""-e '*/help_texts.py'" can be removed soon because newly generated help_texts.py files now contain "# yapf: disable"
    cmd = "yapf -rip -e '*/help_texts.py' -e '*.tpl.py' -e '*/static_root/*' -e '*/templates/*' ."
    ctx.run(cmd, pty=True)


@task(name='sdist')
def setup_sdist(ctx):
    "Create a source distribution."
    atelier.current_project.load_info()
    info = atelier.current_project.SETUP_INFO
    if not info.get('version'):
        return
    if not info.get('name'):
        return
    show_pypi_status(ctx, False)
    dist_dir = str(ctx.sdist_dir).format(prj=info.get('name'))
    args = [sys.executable, "setup.py"]
    args += ["sdist", "--formats=gztar"]
    args += ["--dist-dir", dist_dir]
    ctx.run(' '.join(args), pty=True)


@task(name='release',
      help={
          'batch': "Don't ask for confirmations.",
          'branch': "Create a version branch."
      })
def pypi_release(ctx, batch=False, branch=False):
    """
    Publish a new version to PyPI.

    See https://atelier.lino-framework.org/invlib.html#command-inv-release
    for details.

    """
    atelier.current_project.load_info()
    info = atelier.current_project.SETUP_INFO
    if not info.get('version'):
        return
    version = info['version']
    # dist_dir = Path(ctx.sdist_dir) / info['name']
    dist_dir = str(ctx.sdist_dir).format(prj=info.get('name'))
    dist_dir += "/{name}-{version}.tar.gz".format(**info)

    show_revision_status(ctx)
    show_pypi_status(ctx, True)

    if not (batch or confirm(RELEASE_CONFIRM % info)):
        return

    # args = [sys.executable, "setup.py"]
    # args += ["sdist", "--formats=gztar"]
    # args += ["--dist-dir", dist_dir]
    # args += ["upload"]
    # sdist_cmd = ' '.join(args)
    args = ["twine", "upload"]
    # args +=["--repository-url",""]
    args += [dist_dir]
    sdist_cmd = ' '.join(args)
    if ctx.revision_control_system == 'git' and branch:
        msg = "You might want to ignore this and manually run:\n{}".format(
            sdist_cmd)
        tag_name = "v{}".format(version)
        args = ["git", "branch", tag_name]
        # args = ["git", "tag"]
        # args += ["-a", tag_name]
        # args += ["-m", "'Release %(name)s %(version)s.'" % info]
        res = ctx.run(' '.join(args), pty=True, warn=True)
        if res.exited:
            print(msg)
            return
        args = ["git", "push", "origin", tag_name]
        res = ctx.run(' '.join(args), pty=True, warn=True)
        if res.exited:
            print(msg)
            return

    # pypi_register(ctx)
    ctx.run(sdist_cmd, pty=True)


@task(name='test_sdist')
def test_sdist(ctx):
    """Install a previously created sdist into a temporary virtualenv and
    run test suite.

    """
    info = atelier.current_project.SETUP_INFO
    if not info.get('version'):
        return
    # from atelier.projects import load_projects
    # projects = [p for p in load_projects() if p.SETUP_INFO.get('version') and p['name'] != info['name']]

    with cd(ctx.root_dir):
        ctx.run("rm -Rf tmp/tmp", pty=True)
        ctx.run("virtualenv tmp/tmp", pty=True)
        activate = ". tmp/tmp/bin/activate"

        def vrun(cmd):
            cmd = activate + ';' + cmd
            ctx.run(cmd, pty=True)

        vrun("pip install --download {0} {1}".format(ctx.pypi_dir,
                                                     info['name']))
        # DEPRECATION: pip install --download has been deprecated and will be removed in the future. Pip now has a download command that should be used instead.

        # vrun("pip download {0}".format(info['name']))

        dist_dir = str(ctx.sdist_dir).format(prj=info.get('name'))

        vrun(
            "pip install --no-allow-external --no-index --no-cache-dir -f {} -f {} {}"
            .format(sdist_dir, ctx.pypi_dir, info['name']))
        # vrun("pip install -f {0} {1}".format(ctx.sdist_dir, info['name'])

        vrun("inv test")


@task(name='mm')
def make_messages(ctx):
    "Extract messages, then initialize and update all catalogs."
    extract_messages(ctx)
    init_catalog_code(ctx)
    update_catalog_code(ctx)
    for tree in atelier.current_project.get_doc_trees():
        tree.make_messages(ctx)

    # if False:
    #     pass
    # extract_messages_userdocs()
    # setup_babel_userdocs('init_catalog')
    # setup_babel_userdocs('update_catalog')


@task(name='register')
def pypi_register(ctx):
    """Register this project (and its current version) to PyPI. """
    args = [sys.executable, "setup.py"]
    args += ["register"]
    ctx.run(' '.join(args), pty=True)


@task(name='ci')
def checkin(ctx, today=None):
    """Checkin and push to repository, using today's blog entry as commit message."""

    if ctx.revision_control_system is None:
        return

    if ctx.revision_control_system == 'git':
        from git import Repo
        repo = Repo(ctx.root_dir)
        if not repo.is_dirty():
            print("No changes to commit in {0}.".format(ctx.root_dir))
            return

    show_revision_status(ctx)

    today, now = get_current_date(today)
    entry = get_blog_entry(ctx, today)
    if not entry.path.exists():
        quit("%s does not exist!" % entry.path.absolute())

    msg = entry.url

    if not confirm("OK to checkin %s %s?" % (ctx.project_name, msg)):
        return
    # ~ puts("Commit message refers to %s" % entry.absolute())

    if ctx.revision_control_system == 'hg':
        args = ["hg", "ci"]
    else:
        args = ["git", "commit", "-a"]
    args += ['-m', msg]
    cmd = ' '.join(args)
    ctx.run(cmd, pty=True)
    if ctx.revision_control_system == 'hg':
        ctx.run("hg push %s" % ctx.project_name, pty=True)
    else:
        ctx.run("git push", pty=True)


@task(name='blog')
def edit_blog_entry(ctx, today=None):
    """Edit today's blog entry, create an empty file if it doesn't yet exist.

    :today: Useful when a working day lasted longer than midnight, or
            when you start some work in the evening, knowing that you
            won't commit it before the next morning.  Note that you
            must specify the date using the YYYYMMDD format.

            Usage example::

                $ inv blog -t 20150727

            When :env:`multiple_blog_entries_per_day` is `True`::

                $ inv blog -t 20150727:1232

    """
    if not ctx.editor_command:
        raise MissingConfig("editor_command")
    today, now = get_current_date(today)
    if not ctx.multiple_blog_entries_per_day:
        now = None
    entry = get_blog_entry(ctx, today, now)
    if not entry.path.exists():
        if ctx.languages is None:
            # txt = today.strftime(ctx.long_date_format)
            lng = 'en'
        else:
            lng = ctx.languages[0]
        txt = format_date(today, format='full', locale=lng)
        txt = txt[0].upper() + txt[1:]  # estonian weekdays
        if now is not None:
            txt += " ({})".format(now.strftime("%H:%M"))
            today = str(today) + " " + now.strftime("%H:%M")
        content = ":date: {0}\n\n".format(today) + rstgen.header(1, txt)

        msg = "{0}\nCreate file {1}?".format(content, entry.path)
        if not confirm(msg):
            return
        # for every year we create a new directory.
        yd = entry.path.parent
        if not yd.exists():
            if not confirm("Happy New Year! Create directory %s?" % yd):
                return
            yd.mkdir()
            txt = ".. blogger_year::\n"
            (yd / 'index.rst').write_text(txt)
            set_times(yd.parent / 'index.rst')

        entry.path.write_text(content)
        # touch the year's index for Sphinx:
        # Touch the main index.rst in case it contains a blogger_latest directive.
        # We assume that if there is one, it is there.
        indexes = []
        indexes.append(entry.path.parent / 'index.rst')
        indexes.append(entry.path.parent.parent / 'index.rst')
        indexes.append(entry.path.parent.parent.parent / 'index.rst')
        for i in indexes:
            if i.exists():
                set_times(i)

        # set_times(entry.path.parent.parent / 'index.rst')
        # set_times(Path(ctx.blog_root) / 'index.rst')

    args = [ctx.editor_command.format(entry.path)]
    args += [str(entry.path)]
    # raise Exception("20160324 %s", args)
    ctx.run(' '.join(args), pty=False)


@task(name='pd')
def publish(ctx, only=None):
    """Publish docs. Upload docs to public web server. """
    if not atelier.current_project.published:
        print("Not publishing {}".format(atelier.current_project))
        return
    if not ctx.docs_rsync_dest:
        raise MissingConfig("docs_rsync_dest")

    for tree in atelier.current_project.get_doc_trees():
        if tree.src_path:
            if only is None or tree.rel_path == only:
                tree.publish_docs(ctx)


def show_revision_status(ctx):
    if ctx.revision_control_system == 'hg':
        args = ["hg", "st"]
    elif ctx.revision_control_system == 'git':
        args = ["git", "status"]
    else:
        print("Invalid revision_control_system %r !" %
              ctx.revision_control_system)
        return
    print("-" * 80)
    ctx.run(' '.join(args), pty=True)
    print("-" * 80)


def show_pypi_status(ctx, severe=True):
    """Show project status on PyPI before doing a release.
    """
    info = atelier.current_project.SETUP_INFO
    version = info['version']
    name = info['name']

    assert name and version

    from xmlrpc.client import ServerProxy

    # thanks to https://github.com/pypa/warehouse/issues/8753
    class RateLimitedServerProxy(ServerProxy):

        def __getattr__(self, name):
            time.sleep(1)
            return super(RateLimitedServerProxy, self).__getattr__(name)

    client = RateLimitedServerProxy('https://pypi.python.org/pypi')
    released_versions = client.package_releases(name)
    if len(released_versions) == 0:
        print("No PyPI release of %(name)s has been done so far." % info)
    else:
        urls = client.release_urls(name, released_versions[-1])
        if len(urls) == 0:
            msg = "Last PyPI release was {0} (no files available)."
            msg = msg.format(released_versions[-1])
            print(msg)
        else:
            lastrel = urls[-1]
            # dt = lastrel['upload_time']
            # lastrel['upload_time'] = dt.ISO()
            print(LASTREL_INFO % lastrel)
        if severe and version in released_versions:
            raise Exit("ABORT: %(name)s %(version)s has already been "
                       "released." % info)


def get_blog_entry(ctx, today, now=None):
    """Return an RstFile object representing the blog entry for that date
    in the current project.

    """
    parts = ['blog', str(today.year)]
    if now is None:
        parts.append(today.strftime("%m%d"))
    else:
        parts.append(today.strftime("%m%d") + "_" + now.strftime("%H%M"))
    return RstFile(Path(ctx.blog_root), ctx.blogref_url, parts)


def get_locale_dir(ctx):
    locale_dir = ctx.locale_dir
    if locale_dir is None:
        return
    return Path(locale_dir)


# @task(name='bh')
# def build_help_texts(ctx):
#     """Build help_texts.py file for this project."""
#     if not ctx.help_texts_source:
#         return
#     src_dir = ctx.help_texts_source
#     dest_dir = Path(ctx.build_dir_name)
#     args = ["sphinx-build", "-b", "help_texts"]
#     args += [src_dir, dest_dir]
#     cmd = ' '.join(args)
#     ctx.run(cmd, pty=True)


def extract_messages(ctx):
    """Extract messages from source files to `django.pot` file"""
    ld = get_locale_dir(ctx)
    if not ld:
        return
    args = [sys.executable, "setup.py"]
    args += ["extract_messages"]
    args += ["-o", str(ld / "django.pot")]
    cmd = ' '.join(args)
    # ~ must_confirm(cmd)
    ctx.run(cmd, pty=True)


def init_catalog_code(ctx):
    """Create code .po files if necessary."""
    from lino.core.site import to_locale
    ld = get_locale_dir(ctx)
    if not ld:
        return
    for loc in ctx.languages:
        if loc != 'en':
            f = ld / (loc + '/LC_MESSAGES/django.po')
            if f.exists():
                print("Skip {} because file exists.".format(f))
            else:
                args = [sys.executable, "setup.py"]
                args += ["init_catalog"]
                args += ["--domain django"]
                args += ["-l", to_locale(loc)]
                args += ["-d", str(ld)]
                # ~ args += [ "-o" , f ]
                args += ["-i", str(ld / 'django.pot')]
                cmd = ' '.join(args)
                must_confirm(cmd)
                ctx.run(cmd, pty=True)


def update_catalog_code(ctx):
    """Update .po files from .pot file."""
    from lino.core.site import to_locale
    ld = get_locale_dir(ctx)
    if not ld:
        return
    for loc in ctx.languages:
        if loc != ctx.languages[0]:
            args = [sys.executable, "setup.py"]
            args += ["update_catalog"]
            args += ["--domain django"]
            args += [
                "-o",
                str(ld / to_locale(loc) / 'LC_MESSAGES' / 'django.po')
            ]
            args += ["-i", str(ld / "django.pot")]
            args += ["-l", to_locale(loc)]
            cmd = ' '.join(args)
            # ~ must_confirm(cmd)
            ctx.run(cmd, pty=True)


# @task(name='ls')
# def list_projects(ctx, *cmdline_args):
#     """List your projects."""


def git_projects():
    for prj in load_projects():
        prj.load_info()
        if prj.config['revision_control_system'] == 'git':
            yield prj


@task(name='check')
def check_config(ctx):
    """Check configuration integrity."""
    from git import Repo
    import giturlparse  # pip install git-url-parse
    prj = atelier.current_project
    prj.load_info()
    if prj.config['revision_control_system'] != 'git':
        if (prj.root_dir / '.git').exists():
            print(
                "revision_control_system is not 'git', but .git directory exists"
            )
        return
    info = prj.SETUP_INFO
    if len(info) == 0:
        # print("No SETUP_INFO")
        return
    repo = Repo(prj.root_dir)
    p = giturlparse.parse(repo.remotes.origin.url)
    origin = "https://{p.resource}/{p.owner}/{p.name}".format(p=p)
    if not 'url' in info:
        print("No url in SETUP_INFO (expected: {})".format(origin))
        return
    if info['url'] != origin:
        print("url in SETUP_INFO ({}) differs from git origin {}".format(
            info['url'], origin))
    # srcref_url = getattr(prj.main_package, 'srcref_url', None)
    # if srcref_url and not srcref_url.startswith(origin):
    #     print("srcref_url of {} ({}) must start with git origin ({})".format(
    #         prj.main_package, srcref_url, origin))

    doc_trees = prj.get_xconfig('doc_trees')
    intersphinx_urls = prj.get_xconfig('intersphinx_urls')
    for dtname in doc_trees:
        # public_url = add_slash(prj.get_public_docs_url(dtname))
        if intersphinx_urls:
            for k, url in intersphinx_urls.items():
                checkurl(url, "intersphinx_urls['{}']".format(k))
            # if public_url:
            #     url2 = add_slash(intersphinx_urls[dtname])
            #     if public_url != url2:
            #         print("public_url ({}) differs from main intersphinx url ({})".format(
            #             public_url, url2))
    public_url = prj.get_public_docs_url('docs')
    if not public_url and prj.config['selectable_languages']:
        print("selectable_languages but no public_url")

    url = prj.config.get('blogref_url', None)
    if url:
        r = requests.get(url, allow_redirects=False)
        if r.status_code != 200:
            print("blogref_url ({}) returns {}".format(url, r.status_code))


@task(name='ct')
def commited_today(ctx, today=None):
    """Print all today's commits to stdout."""
    from git import Repo

    list_options = dict()
    if True:
        today, now = get_current_date(today)
        ONEDAY = timedelta(days=1)
        yesterday = today - ONEDAY
        tomorrow = today + ONEDAY
        list_options.update(after=yesterday.strftime("%Y-%m-%d"),
                            before=tomorrow.strftime("%Y-%m-%d"))
    if False:
        list_options.update(max_count=5)

    rows = []

    def load(prj):

        repo = Repo(prj.root_dir)
        it = list(repo.iter_commits(**list_options))
        if len(it) == 0:
            # print("20160816 no commits in {}".format(prj.nickname))
            return

        def fmtcommit(c):

            url = repo.remotes.origin.url
            if url.startswith("git@github.com"):
                url = "https://github.com/" + url[15:-4] \
                      + "/commit/" + c.hexsha
            elif url.startswith("git+ssh://git@github.com"):
                url = "https://github.com/" + url[25:-4] \
                      + "/commit/" + c.hexsha

            s = "`{0} <{1}>`__".format(c.hexsha[-7:], url)
            # if c.message and not c.message.startswith("http://"):
            s += "\n({})".format(c.message.strip())
            return s

        # url = prj.SETUP_INFO.get('url', "oops")
        # desc = "`%s <%s>`__" % (prj.name, url)
        desc = "*{}*".format(prj.nickname)

        for c in it:
            # ts = time.strftime("%H:%M", time.gmtime(c.committed_date))
            ts = time.strftime("%Y-%m-%d %H:%M",
                               time.localtime(c.committed_date))
            rows.append([ts, desc, fmtcommit(c)])

    for p in git_projects():
        load(p)

    rows.sort(key=lambda a: a[0])
    print(rstgen.ul(["{0} in {1}:\n{2}".format(*row) for row in rows]))
    # print rstgen.table(headers, rows)


# @task(name='pull')
# def git_pull(ctx):
#     """Run git pull if it is a git project."""
#     from git import Repo
#     for p in git_projects():
#         with cd(p.root_dir):


def run_in_demo_projects(ctx,
                         py_cmd,
                         cov=False,
                         bare=False,
                         after=None,
                         start=None):
    """
    Run the given Python command line `py_cmd` in each demo project.

    When :attr:`bare` is True, return a string concatenated from the bare output
    in each demo project.

    See also :envvar:`demo_projects`.
    """
    outputs = []
    if after is None and start is None:
        demo_projects = list(ctx.demo_projects)
    else:
        demo_projects = []
        skipping = True
        for dpname in ctx.demo_projects:
            if after and dpname.endswith(after):
                skipping = False
                continue
            if start and dpname.endswith(start):
                skipping = False
            if skipping:
                continue
            demo_projects.append(dpname)

    basename = commonprefix(demo_projects)
    # print("All demo projects start with", basename)
    chop = len(basename)

    if len(demo_projects) == 0:
        return "" if bare else None
    if bare:
        print("Run `{}` in {} demo projects:".format(py_cmd,
                                                     len(demo_projects)),
              end=" ",
              flush=True)
    for dpname in demo_projects:
        if bare:
            print(dpname[chop:], end=" ", flush=True)
        else:
            print("-" * 80)
            print("Run `{}` in {} :".format(py_cmd, dpname))
        dpmodule = import_module(dpname)
        pth = Path(dpmodule.__file__).parent
        # join each demo project to root_dir to avoid failure when
        # `inv prep` is invoked from a subdir of root.
        with cd(pth):
            if cov:
                cmd = "coverage run --append " + py_cmd
                datacovfile = ctx.root_dir / '.coverage'
                if not datacovfile.exists():
                    print('No .coverage file in {0}'.format(ctx.project_name))
                os.environ['COVERAGE_FILE'] = datacovfile
            else:
                cmd = sys.executable + ' ' + py_cmd
            if bare:
                out_stream = io.StringIO()
                result = ctx.run(cmd, out_stream=out_stream)
                if result.exited:
                    print(out_stream.getvalue())
                else:
                    outputs.append(out_stream.getvalue())
            else:
                result = ctx.run(cmd, pty=True)
            if result.exited is None:
                raise Exception("Subprocess did not exit")
            if result.exited:
                raise Exception("Subprocess exited with {}".format(
                    result.exited))
    if bare:
        print("DONE")
        return "\n".join(outputs)


@task(
    name='install',
    help=dict(
        list="Don't install anything, just list the requirements to `stdout`.",
        batch="Don't ask for confirmations."))
def install(ctx, list=False, batch=False, after=None, start=None):
    """
    Install required Python packages to your Python environment and/or
    to your `requirements-install.txt` file.

    See https://atelier.lino-framework.org/invlib.html#command-inv-install
    """

    print("Collect list of required packages from all demo projects.")
    cmd = 'manage.py install --list'
    output = run_in_demo_projects(ctx,
                                  cmd,
                                  bare=True,
                                  after=after,
                                  start=start)
    reqs = sorted(set(output.split()))
    if list:
        print("\n".join(reqs))
        return

    cwd = os.getcwd()

    if batch or confirm("OK to modify your requirements files?"):
        with open(cwd + '/requirements-install.txt', 'w') as f:
            f.write("# this file was generated by `inv install`\n")
            f.write("\n".join(reqs))
        with open(cwd + '/requirements.txt', 'r+') as f:
            content = f.read()
            include = '-r requirements-install.txt'
            if include not in content:
                content = include + '\n' + content
                f.seek(0)
                f.write(content)

    cmd = sys.executable + " -m pip install --upgrade pip {}".format(
        ' '.join(reqs))
    if batch or confirm("{} (y/n) ?".format(cmd)):
        ctx.run(cmd)


@task(
    name='prep',
    help=dict(
        after=
        "Run demo_prep_command only in the demo projects *after* the named one.",
        start=
        "Run demo_prep_command only in the demo projects *starting at* the named one."
    ))
def prep(ctx, cov=False, after=None, start=None):
    """
    Run preparation tasks that need to run before testing.

    See also the `prep_command` and `demo_prep_command` in your `tasks.py` file.

    See https://atelier.lino-framework.org/invlib.html#command-inv-prep

    """
    # if cov:
    #     covfile = ctx.root_dir / '.coveragerc'
    #     if not covfile.exists():
    #         raise Exception('No .coveragerc file in {0}'.format(
    #             ctx.project_name))
    #     # os.environ['COVERAGE_PROCESS_START'] = covfile
    #     ctx.run('coverage erase', pty=True)

    cmd = ctx.prep_command
    if cmd:
        print("-" * 80)
        print("Run main prep command {0} :".format(cmd))
        ctx.run(cmd, pty=True)
    cmd = ctx.demo_prep_command
    if cmd:
        # run_in_demo_projects(ctx, cmd, cov=cov)
        run_in_demo_projects(ctx, cmd, bare=True, after=after, start=start)
        # run_in_demo_projects(ctx, cmd)


# @task(name='cov', pre=[run_tests])
@task(name='cov')
def run_tests_coverage(ctx, html=True, html_cov_dir='htmlcov'):
    """Run all tests and create a coverage report.

    If there is a directory named :xfile:`htmlcov` in your project's
    `root_dir`, then it will write a html report into this directory
    (overwriting any files without confirmation).

    """
    # covfile = ctx.root_dir / '.coveragerc'
    # if not covfile.exists():
    #     print('No .coveragerc file in {0}'.format(ctx.project_name))
    #     return
    if not 'COVERAGE_PROCESS_START' in os.environ:
        msg = "You must set COVERAGE_PROCESS_START before running `inv cov`!"
        raise Exit(msg)

    # test whether this Python installation is configured for coverage
    from coverage import process_startup
    if getattr(process_startup, "coverage", None) is None:
        try:
            import sitecustomize
            fn = os.path.realpath(sitecustomize.__file__)
            msg = "Please add the following to your {} file:".format(str(fn))
        except ImportError:
            msg = 'Please create a sitecustomize.py file with the following content:'

        msg += """
    try:
        import coverage
    except ImportError:
        pass
    else:
        coverage.process_startup()"""

        msg += "\nSee also https://coverage.readthedocs.io/en/coverage-4.3.4/subprocess.html"
        raise Exit(msg)

    ctx.run('coverage erase', pty=True)
    print("Running {0} in {1} within coverage...".format(
        ctx.coverage_command, ctx.project_name))
    ctx.run('coverage run --parallel-mode {}'.format(ctx.coverage_command),
            pty=True)
    if False:
        ctx.run('coverage combine', pty=True)
        ctx.run('coverage report', pty=True)
        if html:
            pth = ctx.root_dir / html_cov_dir
            print("Writing html report to {}".format(pth))
            ctx.run('coverage html -d {}'.format(pth), pty=True)
            if False:
                ctx.run('open {}/index.html'.format(pth), pty=True)
            print('{}/index.html has been generated.'.format(pth))
        ctx.run('coverage erase', pty=True)



# -*- coding: UTF-8 -*-
# python setup.py test -s tests.BasicTests.test_utils
# Copyright 2009-2020 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""
Defines a series of utility classes and functions.

Deprecated. Rather import each utility from its real place.

"""

# import six
# from six.moves import input
# from builtins import str
# from builtins import object
import re

from pathlib import Path
from pprint import pprint
from rstgen.utils import *

# def get_visual_editor():
#     """Returns the name of the visual editor, usually stored in the
#     `VISUAL` environment variable.  If `VISUAL` is not set, return the
#     value of `EDITOR`.

#     https://help.ubuntu.com/community/EnvironmentVariables

#     """
#     return os.environ.get('VISUAL') or os.environ.get('EDITOR')



# Copyright 2019-2021 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

# 20210501 Not yet used. WIP for migrating from argh to click.

# """The main entry point for the :command:`atelier` command.
# """
#
# import click
#
# from atelier.setup_info import SETUP_INFO
#
# from .foreach import foreach
# from .listapps import listapps
#
# @click.group(help="""
# {}
# See https://atelier.lino-framework.org for more information.
#
# This is atelier version {}.
# """.format(SETUP_INFO['description'], SETUP_INFO['version']))
# def main():
#     pass
#
# main.add_command(foreach)
# main.add_command(listapps)
#
# if __name__ == '__main__':
#     main()



# -*- coding: UTF-8 -*-
# doctest atelier/sheller.py
# Copyright 2016-2021 by Rumma & Ko Ltd.
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""Defines the :class:`Sheller` class for testing bash commands in a doctest.

Others have invented similar things before.
For example `doctest2 <https://pythonhosted.org/doctest2/intro_for_existing.html>`__
by Devin Jeanpierre.
Another alternative is `Clatter
<https://clatter.readthedocs.io/en/latest/readme.html>`__
by Michael Delgado.

"""

import os
import sys
import locale
import types
import datetime
import subprocess
from tempfile import TemporaryDirectory


class Sheller:
    """A sheller is a little helper object to be used in tested documents
for running shell commands and testing their output.

When instantiating a Sheller, you can optionally give a single argument, which
is the directory where the commands will run. When you don't specify a
directory, Sheller creates a temporary directory and has all processes run
there.

Usage example:

>>> import os
>>> from atelier.sheller import Sheller
>>> shell = Sheller(os.path.dirname(__file__))
>>> shell('ls -S *.py')  #doctest: +NORMALIZE_WHITESPACE
projects.py
test.py
jarbuilder.py
sheller.py
setup_info.py
sphinxconf.py
utils.py
__init__.py

In case you wonder: We sort the by size because that's reproducible. Seems that
some locales put underscores before letters while others don't.

    """

    def __init__(self, cwd=None):
        if cwd is None:
            self.temp_dir = TemporaryDirectory()
            cwd = self.temp_dir.name
        self.cwd = cwd

    def __call__(self, cmd, **kwargs):
        """Run the specified shell command `cmd` in a subprocess and print its
        output to stdout. This is designed for usage from within a doctest
        snippet.

        If `cmd` is a multiline string, semicolons are automatically inserted as
        line separators. Line continuation

        One challenge is that we cannot simply use `subprocess.call` because
        `sys.stdout` is handled differently when running inside doctest.

        """
        cmd = [ln for ln in cmd.splitlines() if ln.strip()]
        cmd = ';'.join(cmd)

        if self.cwd:
            kwargs.update(cwd=self.cwd)

        process = subprocess.Popen(cmd,
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True,
                                   stderr=subprocess.STDOUT,
                                   shell=True,
                                   **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        output = output.strip()
        print(output)
        # if(output):
        #     print(output.strip())
        # else:
        #     print("(exit status {})".format(retcode))



# -*- coding: utf-8 -*-
# Copyright 2011-2023 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from pathlib import Path
from rstgen import sphinxconf
from rstgen.utils import srcref_url_template
from rstgen.sphinxconf import interproject
from atelier.projects import get_project_from_path
import atelier
import rstgen


def configure(globals_dict, project=None, **kwargs):

    docs_root = Path(globals_dict['__file__']).parent.absolute()
    if project is None:
        project = get_project_from_path(docs_root.parent)
    atelier.current_project = project
    project.load_info()

    rstgen.set_config_var(public_url=project.get_public_docs_url('docs'))
    # print("20220710 sphinxconf.configure()", project.get_xconfig('use_dirhtml'))
    rstgen.set_config_var(use_dirhtml=project.get_xconfig('use_dirhtml'))
    rstgen.set_config_var(
        selectable_languages=project.get_xconfig('selectable_languages'))

    sphinxconf.configure(globals_dict, **kwargs)

    globals_dict['html_context'].update(SETUP_INFO=project.SETUP_INFO,
                                        project=project)

    version = project.SETUP_INFO.get('version', None)
    if version:
        globals_dict.update(release=version)
        globals_dict.update(version='.'.join(version.split('.')[:2]))

    if 'name' in project.SETUP_INFO:
        globals_dict['extensions'] += [
            'sphinx.ext.autodoc', 'sphinx.ext.autosummary'
        ]

    if project.main_package:
        root_mod, tpl = srcref_url_template(project.main_package)
        if tpl:
            globals_dict['extlinks'].update(srcref=(tpl, '(source %s)'))

    interproject.configure(globals_dict)



from atelier.invlib import setup_from_tasks

ns = setup_from_tasks(
    globals(),
    "atelier",
    blogref_url="https://luc.lino-framework.org",
    revision_control_system='git',
    # tolerate_sphinx_warnings=True,
    cleanable_files=['docs/api/atelier.*', 'docs/api/rstgen.*'])



# Copyright 2013-2021 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

import os
import subprocess
import argparse

import rstgen
from atelier.projects import load_projects
from argh import dispatch_command, arg, CommandError

SHOW_DOCTREES = False  # True takes about 6 seconds on my computer


@arg('cmd',
     nargs=argparse.REMAINDER,
     help="The command to run on every project.")
@arg('-l',
     '--showlist',
     default=False,
     help='Show list of projects.')
@arg('-d',
     '--dirty',
     default=False,
     dest='dirty',
     help='Process only projects with a dirty git status.')
@arg('-s', '--start', help='Start from that project, skip those before.')
@arg('-a', '--after', help='Start after that project, skip those before.')
@arg('-u', '--until', help='Only until that project, skip those after.')
@arg('-v',
     '--voice',
     help='Speak the result through speakers when terminated.')
@arg('-r', '--reverse', help='Loop in reverse order.')
def per_project(voice=False,
                start=None,
                after=None,
                until=None,
                showlist=False,
                dirty=False,
                reverse=False,
                *cmd):
    """Loop over all projects, executing the given shell command in the
root directory of each project.  See
https://atelier.lino-framework.org/usage.html

    """

    projects = list(load_projects())
    if dirty:
        projects = filter(lambda p: p.get_status().endswith("!"), projects)
    if reverse:
        projects.reverse()
    if showlist:
        headers = [
            'Project',
            # 'Version',
            'Status',
            'URL'
        ]
        if SHOW_DOCTREES:
            headers.append('doctrees')

        def cells(self):
            self.load_info()
            # config = self.inv_namespace.configuration()
            yield self.nickname
            # yield self.SETUP_INFO.get('version', '')
            # yield self.main_package
            yield self.get_status()
            yield self.SETUP_INFO.get('url', None)
            if SHOW_DOCTREES:
                doc_trees = u', '.join(
                    [u"{}".format(t) for t in self.get_doc_trees()])
                yield doc_trees
            # yield ', '.join(config['doc_trees'])

        print(rstgen.table(headers, [tuple(cells(p)) for p in projects]))
        if len(cmd) == 0:
            return

    if len(cmd) == 0:
        raise CommandError("You must specify a command!")

    def saymsg(msg):
        if voice:
            msg = msg.replace("'", "\'")
            cmd = ("espeak", "'{}'".format(msg))
            subprocess.call(cmd)

    skipping = start is not None or after is not None
    for prj in projects:
        if cmd[0] == 'git':
            prj.load_info()
            if prj.config['revision_control_system'] != 'git':
                continue
        if start and prj.nickname == start:
            skipping = False
        if after and prj.nickname == after:
            skipping = False
            continue
        if skipping:
            continue
        if until and prj.nickname == until:
            skipping = True
        print("==== %s ====" % prj.nickname)
        os.chdir(prj.root_dir)
        rv = subprocess.call(cmd, cwd=prj.root_dir)
        if rv:
            msg = "%s ended with error %s in project %s" % (' '.join(cmd), rv,
                                                            prj.nickname)
            saymsg(msg)
            raise CommandError(msg)

    msg = "Successfully terminated `{}` for all projects"
    msg = msg.format(' '.join(cmd))
    saymsg(msg)
    print(msg)


def main():
    dispatch_command(per_project)


if __name__ == '__main__':
    main()
    # main(auto_envvar_prefix='GETLINO')



# Copyright 2021 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

import os
import shutil
import click

from os.path import join
from importlib import import_module

import rstgen

from getlino.utils import KNOWN_REPOS, JINJA_ENV

KNOWN_APPS = [r for r in KNOWN_REPOS if r.settings_module and r.package_name]


@click.command()
@click.option('--rst/--no-rst',
              default=False,
              help="Output as reStructuredText")
@click.pass_context
def listapps(ctx, rst):
    """
    List the Lino applications known by getlino.

    """

    if rst:
        click.echo(".. Generated by `atelier listapps --rst`")
        click.echo(".. _getlino.apps:\n")
        click.echo(rstgen.header(1, "List of known Lino applications"))
        # click.echo("\nThe following applications are supported by :cmd:`getlino startsite`.\n")
    rows = []
    headings = ["Name", "Short description", "Nickname"]
    for r in KNOWN_APPS:
        m = import_module(r.settings_module)
        s = m.Site
        # r: nickname package_name git_repo settings_module front_end
        # print(r.settings_module)
        if rst:
            cells = [
                ":ref:`{s.verbose_name}<{r.nickname}>`".format(**locals()),
                s.description or '', r.nickname
            ]
            rows.append(cells)
        else:
            click.echo(
                "{r.nickname} : {s.verbose_name} : {s.description}".format(
                    **locals()))
            # if s.description:
            #     click.echo("\n" + s.description.strip() + "\n")
            # if r.git_repo:
            #     print("(`Source repository <{r.git_repo}>`__)".format(**locals()))
    if rst:
        click.echo(rstgen.table(headings, rows))
        tpl = JINJA_ENV.get_template("apps_section.rst")
        for r in KNOWN_APPS:
            m = import_module(r.settings_module)
            s = m.Site
            p = import_module(r.package_name.replace("-", "_"))
            public_url = None
            ai = getattr(p, 'ATELIER_INFO', {})
            if 'public_url' in ai:
                public_url = ai['public_url']
            if hasattr(p, 'intersphinx_urls'):
                public_url = p.intersphinx_urls.get('docs', None)
            context = dict(repo=r,
                           package=p,
                           m=m,
                           site=s,
                           rstgen=rstgen,
                           public_url=public_url)
            click.echo(tpl.render(**context))



from setuptools import setup

fn = 'atelier/setup_info.py'
with open(fn, "rb") as fd:
    exec(compile(fd.read(), fn, 'exec'))
# above line is equivalent to the line below, except that it works
# also in Python 3:
# execfile(fn)
if __name__ == '__main__':
    setup(**SETUP_INFO)



# -*- coding: UTF-8 -*-
# Copyright 2011-2023 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""A minimalistic command-line project management.

See :doc:`/usage`.

"""
import os
import sys
from rstgen.utils import cd
# import getopt
# from distutils.errors import DistutilsArgError
# import pkg_resources
from pathlib import Path
from importlib import import_module

from atelier.invlib.utils import SphinxTree

config_files = [
    '~/.atelier/config.py', '/etc/atelier/config.py', '~/_atelier/config.py'
]

_PROJECTS_LIST = []
_PROJECTS_DICT = {}

# class NullDevice():
#     def write(self, s):
#         pass


def load_inv_namespace(root_dir):
    """
    Execute the :xfile:`tasks.py` file of this project and return its
    `ns`.
    """
    # self._tasks_loaded = True

    tasks_file = root_dir / 'tasks.py'
    if not tasks_file.exists():
        return None
        # raise Exception("No tasks.py file in {}".format(root_dir))
        # return

    # print("20180428 load tasks.py from {}".format(root_dir))
    # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    # https://stackoverflow.com/questions/19009932/import-arbitrary-python-source-file-python-3-3
    # fqname = 'atelier.prj_%s' % self.index
    m = dict()
    with cd(root_dir):
        m["__file__"] = str(tasks_file)
        try:
            with open(tasks_file) as f:
                exec(f.read(), m)
        except Exception as e:
            print("Error while loading {} : {}".format(tasks_file, e))
            return
    return m['ns']


def add_project(root_dir, **kwargs):
    """
    To be called from your :xfile:`config.py` file.

    `root_dir` is the name of a directory expected to contain a
    :xfile:`tasks.py`.

    If no `nickname` is specified, the nickname will be the leaf name
    of that directory.

    Returns a :class:`Project` instance describing the project.
    """
    i = len(_PROJECTS_LIST)
    root_dir = Path(root_dir).absolute().resolve()
    if not root_dir.exists():
        raise Exception("Invalid root directory {}".format(root_dir))
    p = Project(i, root_dir, **kwargs)
    _PROJECTS_LIST.append(p)
    _PROJECTS_DICT[root_dir] = p
    return p


def get_project_from_module(modname):
    """Find the project info for the given Python module."""
    m = import_module(modname)
    if m.__file__ is None:
        raise Exception(
            "Invalid module name {} (is it installed?)".format(modname))
    fn = Path(m.__file__)
    prj = get_project_from_path(fn.parent.parent)
    if prj is None:
        # package installed in site-packages without tasks.py file
        root_dir = fn.parent.absolute().resolve()
        prj = _PROJECTS_DICT.get(root_dir)
        if prj is None:
            prj = add_project(root_dir)
            # raise Exception("20191003 dynamically added from {}".format(root_dir))
    # root_dir = Path(m.__file__).parent.parent
    # prj = add_project(root_dir)
    prj.set_main_package(m)
    # assert prj.main_package is not None
    return prj


def get_project_from_nickname(name):
    "Find the project info for the given nickname."
    for p in _PROJECTS_LIST:
        if p.nickname == name:
            return p


def get_project_from_path(root_dir):
    "Find the project info for the given directory."
    root_dir = root_dir.absolute().resolve()
    prj = _PROJECTS_DICT.get(root_dir)
    if prj is None:
        if (root_dir / 'tasks.py').exists():
            return add_project(root_dir)
        # E.g. when the module was instaled via pip
        print("No tasks.py file in {}".format(root_dir))
    return prj


def load_projects():
    for p in _PROJECTS_LIST:
        yield p


class Project(object):
    """Represents a project.

    .. attribute:: main_package

        The main package (a Python module object).

    .. attribute:: index

        An integer representing the sequence number of this project in
        the global projects list.

    .. attribute:: published

        Whether I am the responsible editor of this repository.

        Default value is `True`. You can set this to `False` in your `config.py`
        file using :meth:`set_published`.

        If this is `False`, the :cmd:`inv pd` command will simply do nothing in
        this directory. We use this to fix `#4361
        <https://jane.mylino.net/#/api/tickets/AllTickets/4361>`__.

    .. attribute:: config

        A dict containing the configuration options of this project.
        See :ref:`atelier.prjconf`.

    """
    main_package = None
    published = None
    # srcref_url = None
    # intersphinx_urls = {}
    SETUP_INFO = None
    atelier_info = {}
    config = None
    inv_namespace = None
    _git_status = None
    _doc_trees = None

    def __init__(self, i, root_dir, nickname=None, published=True):
        # , inv_namespace=None, main_package=None):

        self.index = i
        self.root_dir = root_dir
        self.published = published
        #~ self.local_name = local_name
        #~ self.root_dir = Path(atelier.PROJECTS_HOME,local_name)
        self.nickname = nickname or str(self.root_dir.name)
        # self.name = self.nickname  # might change in load_info()
        # self._loaded = False
        # self._tasks_loaded = False
        # print("20180428 Project {} initialized".format(self.nickname))
        #self.main_package = main_package
        #self.inv_namespace = inv_namespace
        self.config = {
            'root_dir': root_dir,
            'build_dir_name': '.build',  # e.g. ablog needs '_build'
            'project_name': str(root_dir.name),
            'locale_dir': None,
            'help_texts_source': None,
            'help_texts_module': None,
            'tolerate_sphinx_warnings': False,
            'cleanable_files': [],
            'revision_control_system': None,
            'default_branch': 'master',
            'apidoc_exclude_pathnames': [],
            'editor_command': os.environ.get('EDITOR'),
            'fixtures_updater': None,
            'prep_command': "",
            'make_docs_command': "",
            'docs_rsync_dest': "",
            'rsync_command': "",
            'test_command':
            "if [ -d tests ]; then python -m unittest discover -s tests; fi",
            'demo_projects': [],
            'demo_prep_command': "manage.py prep --noinput --traceback",
            # 'coverage_command': '{} inv prep test clean --batch bd'.format(pp),
            'coverage_command': '`which invoke` prep test clean --batch bd',
            'languages': None,
            'selectable_languages': None,
            'blog_root': root_dir / 'docs',
            'blogref_url': None,
            'long_date_format': "%Y%m%d (%A, %d %B %Y)",
            'multiple_blog_entries_per_day': False,
            'sdist_dir': root_dir / 'dist',
            'pypi_dir': root_dir / '.pypi_cache',
            'use_dirhtml': False,
            'doc_trees': ['docs'],
            'intersphinx_urls': {},
        }
        # --progress:  show progress
        # --delete: delete files in dest
        # --times:  preserve timestamps
        # --times fails when several users can publish to the same server
        # alternatively. Only the owner  can change the mtime of a
        # directory, other users can't, even if they have write permission
        # through the group.
        self.config.update(
            rsync_command="rsync -e ssh -r --verbose --progress --delete "
            "--times -L --omit-dir-times --exclude .doctrees ./ {dest_url}")

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.root_dir)

    def set_main_package(self, m):
        self.main_package = m

    def set_published(self, v):
        self.published = v

    def set_namespace(self, ns):
        self.inv_namespace = ns
        ns.configure(self.config)
        if self.main_package is None:
            # when no main_package is given, there must be a namespace
            cfg = ns.configuration()
            name = cfg.get('main_package', None)
            # if inv_name != name:
            #     raise Exception("20180428 {} != {}".format(inv_name, name))
            if name:
                # self.doc_trees = None
                # self.name = name
                self.main_package = import_module(name)
                # if self.main_package is None:
                #     raise Exception("Failed to import {}".format(name))

    def load_info(self):
        """
        The project can be:

        - Loaded from a config file: we know only the root_dir

        - instantiated by get_project_from_path() called in
          setup_from_tasks(): we know also the inv_namespace

        - instantiated by get_project_from_module() (by
          sphinxconf.interproject) : we know also the main_package

        A project can have no inv_namespace
        """

        # inv_namespace = self.inv_namespace or load_inv_namespace(
        #     self.root_dir)

        if self.SETUP_INFO is not None:
            # load_info() has been called before
            return

        self.load_setup_file()

        # if self.main_package is None:
        #     self.config.setdefault('doc_trees', ['docs'])
        # else:
        #     self.config.update(main_package=self.main_package.__name__)

        if self.inv_namespace is None:
            ns = load_inv_namespace(self.root_dir)
            if ns is not None:
                self.set_namespace(ns)

    def load_setup_file(self):
        """
        Load the :xfile:`setup.py` file if it exists.
        """
        setup_file = self.root_dir / 'setup.py'
        if not setup_file.exists():
            # print("20180118 no setup.py file in {}".format(root_dir.absolute()))
            self.SETUP_INFO = {}
            return
        g = dict()
        g['__name__'] = 'not_main'
        sys.args = []
        # sys.stdout = sys.stderr = NullDevice()
        cwd = os.getcwd()
        with cd(str(self.root_dir)):
            with open("setup.py") as f:
                code = compile(f.read(), "setup.py", 'exec')
                try:
                    exec(code, g)
                # except (SystemExit, getopt.GetoptError, DistutilsArgError):
                except SystemExit:
                    # contextmanager doesn't restore cwd when we raise an
                    # exception:
                    os.chdir(cwd)
                    raise Exception(
                        "Oops, {} called sys.exit().\n"
                        "Atelier requires the setup() call to be in a "
                        "\"if __name__ == '__main__':\" condition.".format(
                            setup_file))
        info = g.get('SETUP_INFO')
        if info is None:
            raise Exception(
                "Oops, {} doesn't define a name SETUP_INFO.".format(
                    setup_file))
        self.SETUP_INFO = info
        self.atelier_info = g.get('ATELIER_INFO', {})

    def get_status(self):
        # if self.config['revision_control_system'] != 'git':
        # config = self.inv_namespace.configuration()
        self.load_info()
        if self.config['revision_control_system'] != 'git':
            return ''
        if self._git_status is not None:
            return self._git_status
        from git import Repo
        repo = Repo(self.root_dir)
        try:
            s = str(repo.active_branch)
        except TypeError:
            s = "?"
        if repo.is_dirty():
            s += "!"
        self._git_status = s
        return s

    def get_xconfig(self, name, default=None):
        """Return the specified setting from either main module or tasks.py.

        Why do we need this? Shortly, for example a setting like
        :envvar:`intersphinx_urls`: for a Python package we want to read this
        from Python code (even when the package is installed via pip), but
        atelier also needs this, and also for atelier projects that have no
        python package (but a doctree).

        """
        self.load_info()
        if name in self.atelier_info:
            return self.atelier_info[name]
        if self.inv_namespace is None:
            default = self.config.get(name, default)
        else:
            cfg = self.inv_namespace.configuration()
            default = cfg.get(name, default)
        if self.main_package:
            return getattr(self.main_package, name, default)
        return default

    def get_doc_trees(self):
        """
        Yield one DocTree instance for every item of this project's
        :envvar:`doc_trees`.
        """
        if self._doc_trees is not None:
            return self._doc_trees
        # print("20180504 {} get_doc_tree() {}".format(self, self.config))
        # if not hasattr(ctx, 'doc_trees'):
        #     return
        # cfg = self.config
        self._doc_trees = []
        doc_trees = self.get_xconfig('doc_trees')
        if doc_trees is None or len(doc_trees) == 0:
            return self._doc_trees
        selectable_languages = self.get_xconfig('selectable_languages')
        # print("20180504 {} get_doc_tree() {} {}".format(
        #     self, self.main_package, doc_trees))
        if selectable_languages:
            for dtname in doc_trees:
                for i, lng in enumerate(selectable_languages):
                    if i > 0:
                        self._doc_trees.append(
                            SphinxTree(self, dtname, selectable_lang=lng))
                    else:
                        self._doc_trees.append(SphinxTree(self, dtname))
            return self._doc_trees
        for dtname in doc_trees:
            if isinstance(dtname, str):
                self._doc_trees.append(SphinxTree(self, dtname))
            elif isinstance(dtname, tuple):
                # (BUILDER, PATH)
                clparts = dtname[0].split('.')
                cl = import_module(clparts[0])
                for k in clparts[1:]:
                    cl = getattr(cl, k)
                self._doc_trees.append(cl(self, dtname[1]))
            else:
                raise Exception("Invalid item {} in doc_trees".format(dtname))
        return self._doc_trees

    def get_public_docs_url(self, dtname='docs'):
        """Return the URL where the main doctree of this project is published.

        """
        intersphinx_urls = self.get_xconfig('intersphinx_urls', {})
        return intersphinx_urls.get(dtname, "")

        # for dt in self.get_doc_trees():
        #     # print("20210426", self, repr(dt))
        #     dt.load_conf()
        #     if dt.conf_globals is None:
        #         continue
        #     html_context = dt.conf_globals.get('html_context', None)
        #     if html_context:
        #         public_url = html_context.get('public_url')
        #         if public_url:
        #             return public_url


def load_config():
    for fn in config_files:
        fn = os.path.expanduser(fn)
        if os.path.exists(fn):
            # print("Loading config from {}".format(fn))
            with open(fn) as f:
                code = compile(f.read(), fn, 'exec')
                exec(code)
            # print("Loaded {} projects from {}".format(len(_PROJECTS_DICT), fn))


load_config()
